/***
 * $Id: DefaultFTPFileListParser.java,v 1.1.1.1 2001/11/30 03:19:32 dfs Exp $
 *
 * NetComponents Internet Protocol Library
 * Copyright (C) 1997-2002  Daniel F. Savarese
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library in the LICENSE file; if not, write
 * to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA  02111-1307  USA
 ***/

 /***
 * Copyright (C) 2002  IBM Corporation
 *
 * Modified by IBM Corporation August 12, 2002 to provide the ability
 * to specify the format of the file listing returned by the remote
 * FTP server.  This provides compatibility with ftp servers residing
 * in non "en_US" locales.  The functionality is provided by overloading
 * existing methods within DefaultFTPFileListParser, FTPClient, and
 * FTPFileListParser classes to add a String, fileListingFormat, parameter.
 * See the comments preceding the specific methods for further details
 ***/

package com.oroinc.net.ftp;

import java.io.*;
import java.util.*;

/***
 * DefaultFTPFileListParser is the default implementation of
 * <a href="com.oroinc.net.ftp.FTPFileListParser.html"> FTPFileListParser </a>
 * used by <a href="com.oroinc.net.ftp.FTPClient.html"> FTPClient </a>
 * to parse file listings.
 * Sometimes you will want to parse unusual listing formats, in which
 * case you would create your own implementation of FTPFileListParser and
 * if necessary, subclass FTPFile.
 * <p>
 * <p>
 * @author Daniel F. Savarese
 * @see FTPFileListParser
 * @see FTPFile
 * @see FTPClient#listFiles
 ***/

public final class DefaultFTPFileListParser implements FTPFileListParser {


  // end is one beyond end
  private int __charArrayToInt(char[] arr, int start, int end) {
    int value = 0, decimal;
    decimal = 1;
    while(end-- > start) {
      value+=(decimal*(arr[end] - '0'));
      decimal*=10;
    }
    return value;
  }

  private long __charArrayToLong(char[] arr, int start, int end) {
    long value = 0, decimal;
    decimal = 1;
    while(end-- > start) {
      value+=(decimal*(arr[end] - '0'));
      decimal*=10;
    }
    return value;
  }

  private int __skipWhitespace(char[] cToken, int start) {
    while(start < cToken.length && Character.isWhitespace(cToken[start]))
      ++start;
    return start;
  }

  private int __skipDigits(char[] cToken, int start) {
    while(start < cToken.length && Character.isDigit(cToken[start]))
      ++start;
    return start;
  }

  private int __skipNonWhitespace(char[] cToken, int start) {
    while(start < cToken.length && !Character.isWhitespace(cToken[start]))
      ++start;
    return start;
  }

  private int __skipNonWhitespaceToLower(char[] cToken, int start) {
    while(start < cToken.length && !Character.isWhitespace(cToken[start])){
      cToken[start] = Character.toLowerCase(cToken[start]);
      ++start;
    }
    return start;
  }


  /***
   * Parses an FTP server listing entry (a single line) and returns an
   * FTPFile instance with the resulting information.  If the entry could
   * not be parsed, returns null.
   * <p>
   * @param entry  A single line of an FTP server listing with the
   *                 end of line truncated.
   * @return An FTPFile instance representing the file information.  null if
   *         the entry could be parsed, returns null.
   ***/
  public FTPFile parseFTPEntry(String entry) {
    int access, start, end, type, month, year, hour, minutes;
    boolean isDevice;
    Calendar date;
    StringTokenizer tokenizer;
    String sToken;
    char cToken[];
    FTPFile file;

    try {
      cToken = entry.toCharArray();

      file   = new FTPFile();
      file.setRawListing(entry);

      isDevice = (cToken[0] == 'b' || cToken[0] == 'c');

      switch(cToken[0]) {
      case 'd': type = FTPFile.DIRECTORY_TYPE; break;
      case 'l': type = FTPFile.SYMBOLIC_LINK_TYPE; break;
      default:  type = FTPFile.FILE_TYPE; break;
      }

      file.setType(type);

      for(access = 0, start=1; access < 3; access++) {
	// We use != '-' so we avoid having to check for suid and sticky bits
	file.setPermission(access, FTPFile.READ_PERMISSION,
			   (cToken[start++] != '-'));
	file.setPermission(access, FTPFile.WRITE_PERMISSION,
			   (cToken[start++] != '-'));
	file.setPermission(access, FTPFile.EXECUTE_PERMISSION,
			   (cToken[start++] != '-'));
      }

      start = __skipWhitespace(cToken, start);
      end = __skipDigits(cToken, start);
      file.setHardLinkCount(__charArrayToInt(cToken, start, end));

      start = __skipWhitespace(cToken, end);
      end   = __skipNonWhitespace(cToken, start);
      // Get user and group
      file.setUser(new String(cToken, start, end - start));

      start = __skipWhitespace(cToken, end);
      end   = __skipNonWhitespace(cToken, start);
      file.setGroup(new String(cToken, start, end - start));

      // Get size, if block or character device, set size to zero and skip
      // next two tokens.
      if(isDevice) {
	start = __skipWhitespace(cToken, end);
	end   = __skipNonWhitespace(cToken, start);
	start = __skipWhitespace(cToken, end);
	end   = __skipNonWhitespace(cToken, start);
	// Don't explicitly set size because it is zero by default
      } else {
	start = __skipWhitespace(cToken, end);
	end   = __skipDigits(cToken, start);
	file.setSize(__charArrayToLong(cToken, start, end));
      }

      start = __skipWhitespace(cToken, end);
      end   = __skipNonWhitespaceToLower(cToken, start);

      // Get month
      switch(cToken[start]) {
      case 'a':
	if(cToken[end - 1] == 'r')
	  month = 3;
	else
	  month = 7;
	break;
      case 'd':	month = 11; break;
      case 'f': month = 1; break;
      case 'j':
	if(cToken[end - 1] == 'l')
	  month = 6;
	else if(cToken[start + 1] == 'a')
	  month = 0;
	else
	  month = 5;
	break;
      case 'm':
	if(cToken[end - 1] == 'y')
	  month = 4;
	else
	  month = 2;
	break;
      case 'n': month = 10; break;
      case 'o': month = 9; break;
      case 's': month = 8; break;
      default:	month = 0; break;
      }

      // Get day, and store in access
      start = __skipWhitespace(cToken, end);
      end   = __skipDigits(cToken, start);
      access = __charArrayToInt(cToken, start, end);

      start = __skipWhitespace(cToken, end);
      end   = __skipDigits(cToken, start);

      date = Calendar.getInstance();

      try {
	// If token contains a :, it must be a time, otherwise a year
	if(cToken[end] == ':') {
	  year  = date.get(Calendar.YEAR);
	  hour  = date.get(Calendar.MONTH);
	  if(hour < month)
	    --year;
	  hour = __charArrayToInt(cToken, start, end);
	  start = end + 1;
	  end   = __skipDigits(cToken, start);
	  minutes = __charArrayToInt(cToken, start, end);
	} else {
	  // Have to set minutes or compiler will complain not initialized
	  hour = minutes = -1;
	  year = __charArrayToInt(cToken, start, end);
	}

	date.clear();
	date.set(Calendar.YEAR, year);
	date.set(Calendar.MONTH, month);
	date.set(Calendar.DATE, access);

	if(hour != -1) {
	  date.set(Calendar.HOUR, hour);
	  date.set(Calendar.MINUTE, minutes);
	}

      } catch(IllegalArgumentException e) {
	// Do nothing
      }

      file.setTimestamp(date);

      // This is dangerous, but we're going to assume there is only one
      // space after the date.  Most servers seem to use that format and
      // we need to be able to preserve leading spacesin filenames.
      //start = __skipWhitespace(cToken, end);
      start = end + 1;
      end   = __skipNonWhitespace(cToken, start);

      if(end >= cToken.length) {
	file.setName(new String(cToken, start, end - start));
	return file;
      }

      // Now we have to deal with the possibilities of symbolic links and
      // filenames with spaces.  The filename and/or link may contain
      // spaces, numbers, or appear like the date entry, group, etc.,

      sToken = new String(cToken, start, cToken.length - start);

      if(type == FTPFile.SYMBOLIC_LINK_TYPE) {
	end = sToken.indexOf(" -> ");
	// Give up if no link indicator is present
	if(end == -1) {
	  file.setName(sToken);
	  return file;
	}

	file.setName(sToken.substring(0, end));
	file.setLink(sToken.substring(end + 4));
	return file;
      }

      // For other cases, just take the entire token

      file.setName(sToken);
    } catch(ArrayIndexOutOfBoundsException e) {
      return null;
    } catch(StringIndexOutOfBoundsException e) {
      return null;
    }

    return file;
  }


  /***
   * Parses an FTP server file listing and converts it into a usable format
   * in the form of an array of <code> FTPFile </code> instances.  If the
   * file list contains no files, <code> null </code> is returned, otherwise
   * an array of <code> FTPFile </code> instances representing the files in
   * the directory is returned.
   * <p>
   * @param listStream The InputStream from which the file list should be
   *        read.
   * @return The list of file information contained in the given path.  null
   *     if the list could not be obtained or if there are no files in
   *     the directory.
   * @exception IOException  If an I/O error occurs reading the listStream.
   ***/
  public FTPFile[] parseFileList(InputStream listStream) throws IOException {
    String line;
    Vector results;
    BufferedReader reader;
    FTPFile entry;

    reader = new BufferedReader(new InputStreamReader(listStream));

    if((line = reader.readLine()) == null) {
      results = null;
    } else {
      results = new Vector();

      // This is to handle a line at the beginning of the listing
      // that says "total xx" or "Gemstat xx" or something else.
      if(line.toLowerCase().startsWith("total"))
	line = reader.readLine();
      else {
	if((entry = parseFTPEntry(line)) != null)
	  results.addElement(entry);
	line = reader.readLine();
      }

      while(line != null) {
	if(line.length() == 0 || (entry = parseFTPEntry(line)) == null) {
	  results = null;
	  break;
	}
	results.addElement(entry);
	line = reader.readLine();
      }
    }

    // Finish reading from stream just in case
    if(line != null)
      while((line = reader.readLine()) != null);

    reader.close();

    if(results != null) {
      FTPFile[] result;

      result = new FTPFile[results.size()];
      if(result.length > 0)
	results.copyInto(result);
      return result;
    }

    return null;
  }

    // **************** Following are functions added for CR 13518 *****************

   /***
   * * Added by IBM Corporation
   * The existing listFiles() methods support only locale 'en_US' and
   * the FTP operation fails when the FTP server is running in locale
   * other than 'en_US' (say 'en_GB', etc..)
   * This method has been added to support other locales.
   *
   * Parses an FTP server file listing and converts it into a usable format
   * in the form of an array of <code> FTPFile </code> instances.  If the
   * file list contains no files, <code> null </code> is returned, otherwise
   * an array of <code> FTPFile </code> instances representing the files in
   * the directory is returned.
   * <p>
   * @param listStream The InputStream from which the file list should be
   *        read.
   * @param ftpFormatString the format in which the directory is listed while
   *         executing LIST commands. (Eg1. if the result of LIST command is something like
   *         drwxr-xr-x  15 admin  clearuse     1024 Feb 12 17:31 CrossWorlds
   *         then the value of this param should be "P:L:U:G:S:D:M:T:N"
   *
   *         Eg2. if the result of LIST command is something like
   *         drwxr-xr-x  15 admin  clearuse     1024 12 Feb 17:31 CrossWorlds
   *         then the value of this param should be "P:L:U:G:S:M:D:T:", where
   *         P - for permission
   *         L - for links
   *         U - for user
   *         G - for group
   *         S - for size
   *         D - for date
   *         M - for month
   *         T - for time
   *         N - for name
   * @return The list of file information contained in the given path.  null
   *     if the list could not be obtained or if there are no files in
   *     the directory.
   * @exception IOException  If an I/O error occurs reading the listStream.
   ***/

  public FTPFile[] parseFileList(InputStream listStream , String ftpFormatString) throws IOException {

    String line;
    Vector results;
    BufferedReader reader;
    FTPFile entry;

    reader = new BufferedReader(new InputStreamReader(listStream));

    if((line = reader.readLine()) == null) {
      results = null;
    } else {
      results = new Vector();

      // This is to handle a line at the beginning of the listing
      // that says "total xx" or "Gemstat xx" or something else.
      if(line.toLowerCase().startsWith("total"))
	line = reader.readLine();
      else {
	if((entry = parseFTPEntry(line,ftpFormatString)) != null)
	  results.addElement(entry);
	line = reader.readLine();
      }

      while(line != null) {
	if(line.length() == 0 || (entry = parseFTPEntry(line,ftpFormatString)) == null) {
	  results = null;
	  break;
	}
	results.addElement(entry);
	line = reader.readLine();
      }
    }

    // Finish reading from stream just in case
    if(line != null)
      while((line = reader.readLine()) != null);

    reader.close();

    if(results != null) {
      FTPFile[] result;

      result = new FTPFile[results.size()];
      if(result.length > 0)
	results.copyInto(result);
      return result;
    }

    return null;
  }

   /***
   * * Added by IBM Corporation
   * The existing listFiles() methods support only locale 'en_US' and
   * the FTP operation fails when the FTP server is running in locale
   * other than 'en_US' (say 'en_GB', etc..)
   * This method has been added to support other locales.
   *
   * Parses an FTP server listing entry (a single line) and returns an
   * FTPFile instance with the resulting information.  If the entry could
   * not be parsed, returns null.
   * <p>
   * @param entry  A single line of an FTP server listing with the
   *                 end of line truncated.
   * @param ftpFormatString the format in which the directory is listed while
   *         executing LIST commands. (Eg1. if the result of LIST command is something like
   *         drwxr-xr-x  15 admin  clearuse     1024 Feb 12 17:31 CrossWorlds
   *         then the value of this param should be "P:L:U:G:S:M:D:T:N"
   *
   *         Eg2. if the result of LIST command is something like
   *         drwxr-xr-x  15 admin  clearuse     1024 12 Feb 17:31 CrossWorlds
   *         then the value of this param should be "P:L:U:G:S:D:M:T:N", where
   *         P - for permission
   *         L - for links
   *         U - for user
   *         G - for group
   *         S - for size
   *         D - for date
   *         M - for month
   *         T - for time
   *         N - for name
   * @return An FTPFile instance representing the file information.  null if
   *         the entry could be parsed, returns null.
   ***/
  public FTPFile parseFTPEntry(String entry, String ftpFormatString) throws IOException{

  StringTokenizer formatStringTokenizer = new StringTokenizer(ftpFormatString,":");
  String newEntry = null;
  String tokenArray[] = new String[formatStringTokenizer.countTokens()];
  FTPFile file;
  String parsedString;
  int cntr = 0;
  // CR 30237 start
  String timeStr = "";
  // CR 30237 end
  //Tokenize formatSting 'P:L:U:G:S:M:D:T:N'
  while(formatStringTokenizer.hasMoreTokens())
     tokenArray[cntr++] = formatStringTokenizer.nextToken(":");

   HashMap patternTable = mapFormatStringWithListing(tokenArray,entry);

  //form the expected pattern P:L:U:G:S:M:D:T:N
  if(patternTable.get("P") != null)
    parsedString = new String(patternTable.get("P").toString());
  else
    throw new IOException("Unable to locate the 'Permission' part of 'ls -l'");
  if(patternTable.get("L") != null)
     parsedString += patternTable.get("L").toString();
  else
     throw new IOException("Unable to locate the 'No. of Hard links' part of 'ls -l'");
  if(patternTable.get("U") != null)
       parsedString += patternTable.get("U").toString();
  else
    throw new IOException("Unable to locate the 'User name' part of 'ls -l'");
  if(patternTable.get("G")!= null)
       parsedString += patternTable.get("G").toString();
  else
    throw new IOException("Unable to locate the 'Group name' part of 'ls -l'");
  if(patternTable.get("S")!= null)
       parsedString += patternTable.get("S").toString();
  else
    throw new IOException("Unable to locate the 'Size' part of 'ls -l'");
  if(patternTable.get("M")!= null)
       parsedString += patternTable.get("M").toString();
  else
    throw new IOException("Unable to locate the 'Month' part of 'ls -l'");
  if(patternTable.get("D")!= null)
       parsedString += patternTable.get("D").toString();
  else
    throw new IOException("Unable to locate the 'Date' part of 'ls -l'");
  if(patternTable.get("T")!= null) {
	   // CR 30237 start
	   // if hour and minute is separated by a character other than colon, replace it with colon
	   timeStr = (String)patternTable.get("T").toString().trim();
	   if (timeStr.charAt(2) != ':') {
		   parsedString += patternTable.get("T").toString().replace (timeStr.charAt(2), ':');
	   }
	   else {
		   parsedString += patternTable.get("T").toString();
	   }
       // CR 30237 end
   }
  else
    throw new IOException("Unable to locate the 'Time' part of 'ls -l'");
  if(patternTable.get("N")!= null)
       parsedString += patternTable.get("N").toString();
  else
    throw new IOException("Unable to locate the 'File Name' part of 'ls -l'");
  //parsedString is in the expected format 'P:L:U:G:S:M:D:T:N'
  file = parseFTPEntry(parsedString);
   return file;
  }

 /***
 * * Added by IBM Corporation
 * The existing listFiles() methods support only locale 'en_US' and
 * the FTP operation fails when the FTP server is running in locale
 * other than 'en_US' (say 'en_GB', etc..)
 * This method has been added to support other locales.
 *
 * This function stores the tokens in formatString (formatString is something like
 * "P:L:U:G:S:M:D:T:N" )as the key in the hastable and the tokens from the result of
 * LIST command (something like "drwxr-xr-x  15 admin  clearuse     1024 Feb 12 17:31 CrossWorlds"
 * as value in hastable and returns the hastable.
 * For the above example, the hastable will look like
 *  Key          Value
 *  P            drwxr-xr-x
 *  L            15
 *  U            admin
 *  G            clearuse
 *  S            1024
 *  M            Feb
 *  D            12
 *  T            17:31
 *  N            CrossWorlds
 ***/
 private HashMap mapFormatStringWithListing(String[] formatPattern,String entry) throws IOException
 {
     char cToken[];
     cToken = entry.toCharArray();
     String tokenList[]= new String[formatPattern.length];
     int start = 0, end = 0, cntr = 0,tmp = 0;
     HashMap patternTable = new HashMap(formatPattern.length);

     end = this.__skipNonWhitespace(cToken,start);
     tokenList[cntr++] = entry.substring(start,end);
     start = tmp = end;

      while(start < entry.length())
     {
       while(tmp < entry.length() && Character.isWhitespace(cToken[tmp]))
           tmp++;
       end = tmp;
       while(end < entry.length() &&!Character.isWhitespace(cToken[end]))
           end++;
       //Update token list with 'ls -l' entry.
       tokenList[cntr++] = entry.substring(start,end);
       start = tmp = end;
       //end = 0;
     }
     // check whether the no. of token in formatSTring = no. of token in listing
     //else throw exception saying formatstring is not in proper manner
     if(formatPattern.length != tokenList.length )
        throw new IOException("FormatString doesn't match with the actual 'ls -l' listing." +
                              "Change the formatString and try again");
     //fill patternTable with'P' = '-rw-rw-r--' etc.
     for(int i=0; i<formatPattern.length ;i++)
         patternTable.put(formatPattern[i],tokenList[i]);

     return patternTable;
 }

}
